

/*  TOONSAVE.C - ScreenSaverProc(), RegisterDialogClasses(), 
 *   ScreenSaverConfigureDialog() and other support code for 
 *   wintoon screen saver.            
 *
 *   ToonSave is a sample screen saver application. It moves a cartoon
 *   around on the display 
 *
 *    (C) Copyright Microsoft Corp. 1991, 1994.  All rights reserved.
 *
 *    You have a royalty-free right to use, modify, reproduce and 
 *    distribute the Sample Files (and/or any modified version) in 
 *    any way you find useful, provided that you agree that 
 *    Microsoft has no warranty obligations or liability for any 
 *    Sample Application Files which are modified. 
 */

#include <windows.h> 
#include <windowsx.h> 
#include <stdlib.h>
#include <time.h>
#include <string.h>
#include <commdlg.h>
#include <mmsystem.h>
#include <digitalv.h>
#include <memory.h>
#include <dib.h>
#include <wintoon.h>
#include "toonsave.h"
#include <mciavi.h>


// Toon (avi) name is hardwired in as app name
// in OnCreate. This is used to contain the
// full \path\filename.ext of avi
char far szMovie[_MAX_PATH+1];

// Define # of frames to play
DWORD dwFrom = 0;
DWORD dwTo = 519;


// Global used by SCRNSAVE.LIB. Required for all screen savers.
char szAppName[40];

// Globals specific to screen saver.
BYTE iTransparent = 0;
BOOL fRedrawBackground = FALSE;
char szSoundName[] = "Sound";
char szName[]="Display a Toon";
WORD hMovie = 0;
char szBuf[128];
PDIB pdib= NULL;
RECT rcWnd;


/* Externals defined in SCRNSAVE.LIB. Required for all screen savers.
 */
HINSTANCE _cdecl hMainInstance = 0;
HWND _cdecl hMainWindow = 0;
char _cdecl szName[TITLEBARNAMELEN];
char _cdecl szIsPassword[22];
char _cdecl szIniFile[MAXFILELEN];
char _cdecl szScreenSaver[22];
char _cdecl szPassword[16];
char _cdecl szDifferentPW[BUFFLEN];
char _cdecl szChangePW[30];
char _cdecl szBadOldPW[BUFFLEN];
char _cdecl szHelpFile[MAXFILELEN];
char _cdecl szNoHelpMemory[BUFFLEN];
UINT _cdecl MyHelpMessage = 0;
HOOKPROC _cdecl fpMessageFilter;


HBITMAP hbmImage = 0;   	            // image handle
WORD wElapse = 5000;	                // speed parameter - how often to restart toon
WORD wTimer = 0;                        // timer id
int xPos = 0;                           // current x position
int yPos = 0;                           // current y position
BOOL bSound = 0;                        // sound on/off flag
BOOL bPassword = 0;                     // password protected?


#define WM_TOONDRAW	(WM_USER + 1)


// Forward references
BOOL OnCreate(HWND, CREATESTRUCT FAR*);
void OnToonDraw(HWND, HTOON, LONG);
BOOL OnStop(HWND);
BOOL OnPlay(HWND);
BOOL OnOpenFile(HWND, LPSTR);
void OnDestroy(HWND);
int GetAVIColors(HWND, HTOON, char far*, LPRGBQUAD);


/* ScreenSaverProc - Main entry point for screen saver messages.
 *  This function is required for all screen savers.
 *
 * Params:  Standard window message handler parameters.
 *
 * Return:  The return value depends on the message.
 *
 *  Note that all messages go to the DefScreenSaverProc(), except
 *  for ones we process.
 */
LONG FAR PASCAL __export ScreenSaverProc(HWND hWnd, UINT msg, WPARAM wParam, LPARAM lParam)
{
   static int XOffset, YOffset;
	DWORD ErrCode;
	MCI_DGV_STATUS_PARMS	mciStatus;
	MCI_DGV_STATUS_PARMS far*	pmciStatus;
    
    RECT rc;
    static WORD wBottomCount;
    switch (msg)
    {
    	case	SC_SCREENSAVE:
    		return FALSE;
	
        case  WM_CREATE:                             // wintoon-specific
        	return HANDLE_WM_CREATE(hWnd, wParam, lParam, OnCreate);
            break;

        case WM_TIMER:                              // screen saver-specific
			if (hMovie)
			{
				mciStatus.dwCallback = NULL;
				mciStatus.dwReturn   = 0;
				mciStatus.dwItem     = MCI_STATUS_MODE;
				mciStatus.dwTrack    = 0;
				pmciStatus = &mciStatus;
				
				ErrCode = mciSendCommand(hMovie, MCI_STATUS, MCI_STATUS_ITEM, (DWORD)pmciStatus);
				if(ErrCode)
				{
					mciGetErrorString(ErrCode, szBuf, sizeof(szBuf));
					MessageBox(hWnd, szBuf, "Status error", MB_ICONEXCLAMATION);
					return FALSE;
				}
				else if(MCI_MODE_PLAY != mciStatus.dwReturn)
				{
					// Move the toon around and then play it again
					SIZE size;
					MCI_DGV_RECT_PARMS mciRect;
					div_t div_result;
					HTOON hToon = ToonFromWindow(hWnd);
					mciSendCommand (hMovie, MCI_WHERE, MCI_WAIT | MCI_DGV_WHERE_SOURCE, (DWORD)(LPMCI_DGV_RECT_PARMS) &mciRect);
					size = ToonSize (hToon);
				    GetClientRect(hWnd, &rc);  
					div_result = div(rand(), (rc.right - mciRect.rc.right));
					xPos = div_result.rem;
					div_result = div(rand(), (rc.bottom - mciRect.rc.bottom));
					yPos = div_result.rem;
					OffsetRect (&(mciRect.rc), xPos, yPos);
					// MCI_PUT doesn't expect a real rect, 
					// it wants a top, left, width & heigth
					mciRect.rc.bottom = size.cy;
					mciRect.rc.right = size.cx; 
					mciSendCommand (hMovie, MCI_PUT, MCI_WAIT | MCI_DGV_PUT_DESTINATION | MCI_DGV_RECT, (DWORD)(LPMCI_DGV_RECT_PARMS) &mciRect);
					fRedrawBackground = TRUE;  // force a re-draw of background
					OnPlay(hWnd);
				}
			}
			break;
		
        case WM_DESTROY:	// wintoon-specific
            /* Destroy any objects we created
             */
			OnDestroy(hWnd);
            break;

		case WM_ERASEBKGND:
			GetClientRect(hWnd,&rc);
			FillRect((HDC)wParam,&rc,(HBRUSH)GetStockObject(WHITE_BRUSH));
			return 0L;
		
		case WM_PAINT:
			return HANDLE_WM_PAINT(hWnd, param, lParam, ToonOnPaint);
			break;

		case WM_PALETTECHANGED:
			return HANDLE_WM_PALETTECHANGED(hWnd, wParam, lParam, ToonOnPaletteChanged);
            break;

		case WM_QUERYNEWPALETTE:
			return HANDLE_WM_QUERYNEWPALETTE(hWnd, wParam, lParam, ToonOnQueryNewPalette);
            break;

		case WM_TOONDRAW:
			OnToonDraw(hWnd, (HTOON)wParam, lParam);
			return TRUE;

	    default:
		    break;
        }
	    return DefScreenSaverProc(hWnd, msg, wParam, lParam);
}

/* RegisterDialogClasses -- Entry point for registering window
 * classes required by configuration dialog box.
 *
 * Params:  hWnd -- Handle to window
 *
 * Return:  None
 */
BOOL RegisterDialogClasses(HINSTANCE hInst)
{
     return TRUE;
}


/* ScreenSaverConfigureDialog -- Dialog box function for configuration
 * dialog.
 *
 * Params:  hWnd -- Handle to window
 *
 * Return:  None
 */
BOOL FAR PASCAL __export ScreenSaverConfigureDialog(HWND hDlg, UINT msg, WPARAM wParam, LPARAM lParam)
{
  static HWND hIDOK;
  static HWND hSetPassword;

    switch (msg)
    {
        case WM_INITDIALOG:                         // wintoon-specific
            GetIniEntries();
            GetIniSettings();
            SendDlgItemMessage(hDlg, ID_SOUND, BM_SETCHECK, bSound, NULL);
            SendDlgItemMessage(hDlg, ID_PASSWORDPROTECTED, BM_SETCHECK,
                bPassword, NULL);
            hSetPassword=GetDlgItem(hDlg, ID_SETPASSWORD);
            EnableWindow(hSetPassword, bPassword);
            hIDOK=GetDlgItem(hDlg, IDOK);
            return TRUE;

        case WM_COMMAND:                            // wintoon-specific
            switch (wParam)
            {
                case IDOK:
                    bSound     = IsDlgButtonChecked(hDlg, ID_SOUND);
                    bPassword  = IsDlgButtonChecked(hDlg, ID_PASSWORDPROTECTED);
                    WriteProfileInt(szAppName, szSoundName, bSound);
                    WriteProfileInt(szAppName, szIsPassword, bPassword);

                    EndDialog(hDlg, TRUE);
                    return TRUE;

                case IDCANCEL:
                    EndDialog(hDlg, FALSE);
                    return TRUE;

                case ID_SETPASSWORD:
                {
                    FARPROC fpDialog;

                    if((fpDialog = MakeProcInstance(DlgChangePassword,hMainInstance)) == NULL)
                        return FALSE;
                    DialogBox(hMainInstance, MAKEINTRESOURCE(DLG_CHANGEPASSWORD), 
                              hDlg, fpDialog);
                    FreeProcInstance(fpDialog);
                    SendMessage(hDlg, WM_NEXTDLGCTL, hIDOK, 1l);
                    break;
                }

                case ID_PASSWORDPROTECTED:
                    bPassword ^= 1;
                    CheckDlgButton(hDlg, wParam, bPassword);
                    EnableWindow(hSetPassword, bPassword);
                    break;

                case ID_HELP:
DoHelp:
#if 0
                    bHelpActive=WinHelp(hDlg, szHelpFile, HELP_CONTEXT, IDH_DLG_WINTOON);
                    if (!bHelpActive)
                        MessageBox(hDlg, szNoHelpMemory, szName, MB_OK);
#else
                    MessageBox(hDlg, "Insert your call to WinHelp() here.",
                        szName, MB_OK);
#endif
                    break;
            }
            break;
        default:
            if (msg==MyHelpMessage)     // Context sensitive help msg.
                goto DoHelp;
    }
    return FALSE;
}


/* GetIniSettings -- Get initial settings from WIN.INI
 *
 * Params:  None
 *
 * Return:  None
 */
static void GetIniSettings(void)
{
    bSound =     GetPrivateProfileInt(szAppName, szSoundName, DEF_SOUND, szIniFile);
    bPassword =  GetPrivateProfileInt(szAppName, szIsPassword, FALSE, szIniFile);
}



/* WriteProfileInt - Write an unsigned integer value to CONTROL.INI.
 *
 * Params:  name - szSection - [section] name in .INI file
 *                 szKey     - key= in .INI file
 *                 i         - value for key above
 *
 * Return:  None
 */
static void WriteProfileInt(LPSTR szSection, LPSTR szKey, int i) 
{
    char achBuf[40];

    // write out as unsigned because GetPrivateProfileInt() can't
    // cope with signed values!
    wsprintf(achBuf, "%u", i);
    WritePrivateProfileString(szSection, szKey, achBuf, szIniFile);
}

/* GetIniEntries -- Load Common Strings from stringtable
 *
 * Params:  None
 *
 * Return:  None
 */
void GetIniEntries(void)
{
  LoadString(hMainInstance, idsIsPassword, szIsPassword, 22);
  LoadString(hMainInstance, idsIniFile, szIniFile, MAXFILELEN);
  LoadString(hMainInstance, idsScreenSaver, szScreenSaver, 22);
  LoadString(hMainInstance, idsPassword, szPassword, 16);
  LoadString(hMainInstance, idsDifferentPW, szDifferentPW, BUFFLEN);
  LoadString(hMainInstance, idsChangePW, szChangePW, 30);
  LoadString(hMainInstance, idsBadOldPW, szBadOldPW, 255);
  LoadString(hMainInstance, idsHelpFile, szHelpFile, MAXFILELEN);
  LoadString(hMainInstance, idsNoHelpMemory, szNoHelpMemory, BUFFLEN);
}



/* DibFlip -- Flip a DIB over
 *
 * Params:  PDIB - Pointer to DIB
 *
 * Return:  BOOL indicating success
 */
BOOL DibFlip(PDIB pdib)
{
	UINT cLine = DibHeight(pdib);
	UINT cFlip = cLine / 2;
	long cbLine = DibWidthBytes(pdib);
	UINT i;
	HPBYTE lpTop, lpBottom, lpTemp;
	static BYTE temp[640];
	
	//
	// Handle stupid cases
	//
	if (cLine < 2)
	{
		pdib->biHeight = -pdib->biHeight;
		return TRUE;
	}

	if (cbLine > 640)
	{
		if ((lpTemp = GlobalAllocPtr(GHND, cbLine)) == NULL)	
			return FALSE;
	}
	else
	{
		lpTemp = &temp[0];
	}

	lpTop = DibPtr(pdib);
	lpBottom = lpTop + (cLine - 1) * cbLine;
	for (i = 0 ; i < cFlip ; i++)
	{
		if (lpBottom < lpTop)
			return FALSE;
		hmemcpy(lpTemp, lpTop, cbLine);
		hmemcpy(lpTop, lpBottom, cbLine);
		hmemcpy(lpBottom, lpTemp, cbLine);

		lpTop += cbLine;
		lpBottom -= cbLine;
	}
	
	if (lpTemp != &temp[0])
		GlobalFreePtr(lpTemp);
	pdib->biHeight = -pdib->biHeight;
	return TRUE;
}


/* OnStop -- Stop the toon
 *
 * Params:  handle to window
 *
 * Return:  BOOL indicating success
 */
BOOL OnStop(HWND hWnd)
{
	MCI_DGV_STOP_PARMS mciStop = { NULL };
	DWORD ErrCode;
	
	if(hMovie)
	{
		ErrCode = mciSendCommand(hMovie, MCI_STOP, 0, (DWORD)(LPMCI_DGV_STOP_PARMS)&mciStop);
		if (ErrCode)
		{
			mciGetErrorString(ErrCode, szBuf, sizeof(szBuf));
			MessageBox(hWnd, szBuf, "Couldn't stop", MB_ICONEXCLAMATION);
			return FALSE;
		}
	}
	return TRUE;
}


/* OnDestroy -- Clean up
 *
 * Params:  handle to window
 *
 * Return:  None
 */
void OnDestroy(HWND hWnd)
{
	MCI_GENERIC_PARMS mciGeneric;
	HTOON hToon;
	DWORD ErrCode;

	if( wTimer )   
		KillTimer(hWnd, ID_TIMER);
	
	OnStop(hWnd);
	if (hMovie)
	{
		ErrCode = mciSendCommand(hMovie, MCI_CLOSE, (DWORD)(LPMCI_GENERIC_PARMS)&mciGeneric, 0);
		if (ErrCode)
		{
			mciGetErrorString(ErrCode, szBuf, sizeof(szBuf));
			MessageBox(hWnd, szBuf, "Couldn't close", MB_ICONEXCLAMATION);
		}
		hMovie = NULL;
	}

	if ((hToon = ToonFromWindow(hWnd)) != NULL)
		ToonDestroy(hToon);

	if (pdib)
	{
		DibFree(pdib);
		pdib = NULL;
	}
	PostQuitMessage(0);
}




/* OnPlay -- Start playing the toon
 *
 * Params:  handle to window
 *
 * Return:  BOOL indicating success
 */
BOOL OnPlay(HWND hWnd)
{
	MCI_DGV_PLAY_PARMS      mciPlay;
	DWORD ErrCode, dwFlags = MCI_FROM | MCI_TO | MCI_MCIAVI_PLAY_WINDOW;
	HCURSOR hcursor;
	HTOON hToon = ToonFromWindow(hWnd);
		
	mciPlay.dwFrom = dwFrom;
	mciPlay.dwTo = dwTo;

	ToonClearRects(hToon);
	hcursor = SetCursor(LoadCursor(NULL, IDC_WAIT));
	ErrCode = mciSendCommand(hMovie, MCI_PLAY, dwFlags, (DWORD)(LPMCI_DGV_PLAY_PARMS)&mciPlay);
	SetCursor(hcursor);
	if (ErrCode)
	{
		mciGetErrorString(ErrCode, szBuf, sizeof(szBuf));
		MessageBox(hWnd, szBuf, "Couldn't play", MB_ICONEXCLAMATION);
		return FALSE;
	}
	return TRUE;
}



/* GetAVIColors -- Get the toon's colors
 *
 * Params:  handle to window, handle to toon, name of movie, pointer to quad
 *			color array
 *
 * Return:  BOOL indicating success
 */
int GetAVIColors(HWND hWnd, HTOON hToon, char far *szMovie, LPRGBQUAD vColors)
{
int iStatus = FALSE;
MCI_DGV_STATUS_PARMS mciStatus;
MCI_DGV_STATUS_PARMS far*	pmciStatus;
HDC hdc, hdcmem, hdcold;
HBITMAP hBitMap;
HPALETTE hpal;
PDIB pdib;
DWORD ErrCode;
RECT rcWnd;

hdc = GetDC(hWnd);
if(hdc)
{
	GetClientRect(hWnd, &rcWnd);
	hdcmem = CreateCompatibleDC(hdc);
	hBitMap = CreateCompatibleBitmap(hdc, rcWnd.right, rcWnd.bottom);
	hdcold = SelectObject(hdcmem, hBitMap);
	BitBlt(hdcmem,0, 0, rcWnd.right, rcWnd.bottom, hdc, 0, 0, SRCCOPY);
	pdib = DibFromBitmap(hBitMap, 0, 0, NULL, NULL);
	SelectObject(hdcmem, hdcold);
	DeleteDC(hdcmem);
	DeleteObject(hBitMap);
	ReleaseDC(0, hdc);
	
		if ((hMovie = ToonOpenMovie(hToon, szMovie, iTransparent)) != NULL)
	    {
			mciStatus.dwCallback = NULL;
			mciStatus.dwReturn   = 0;
			mciStatus.dwItem     = MCI_ANIM_STATUS_HPAL;
			mciStatus.dwTrack    = 0;
			pmciStatus = &mciStatus;
			ErrCode = mciSendCommand(hMovie, MCI_STATUS, MCI_STATUS_ITEM, (DWORD)pmciStatus);
			if (ErrCode)
			{
				mciGetErrorString(ErrCode, szBuf, sizeof(szBuf));
				MessageBox(hWnd, szBuf, "Couldn't get pal info", MB_ICONEXCLAMATION);
				return FALSE;
			}
			hpal = (HPALETTE)mciStatus.dwReturn;
			GetPaletteEntries(hpal, 0, 256, (PALETTEENTRY far*)vColors);
			return TRUE;
	    }
	    else
			return FALSE;
	}
	else
		return FALSE;
}


/* OnCreate -- handle WM_CREATE message
 *
 * Params:  Standard OnCreate parameters
 *			
 *
 * Return:  BOOL indicating success
 */
BOOL OnCreate(HWND hWnd, CREATESTRUCT FAR* lpCreate)
{

	HDC hDCMem;
	HTOON hToon;
	MCI_GENERIC_PARMS mciGeneric;
	DWORD ErrCode;
	HDC hDCScreen;
	RGBQUAD Colors[256];
	HGDIOBJ hOldBitmap;
  	int nLen;

	srand( (unsigned)time( NULL ) );
    
	// get path of executable
	GetModuleFileName(lpCreate->hInstance, szMovie, _MAX_PATH);

	nLen = _fstrlen(szMovie)-4;	// Get length of name, with path less length of suffix
	szMovie[nLen] = '\0';		// Get rid of suffix
	lstrcat(szMovie, (char far*)".AVI");	// We now have AVI path & name

	if ((hToon = ToonCreate(hWnd, lpCreate->cx, lpCreate->cy, 0)) == NULL)
		return FALSE;
	ToonSetMessage(hToon, WM_TOONDRAW);
	
	
	/* Load the strings from the STRINGTABLE
	*/
	GetIniEntries();
	
	/* Load the initial bounce settings.
	*/
	GetIniSettings();
	


	if( TRUE == GetAVIColors(hWnd, hToon, szMovie, Colors))
	{
		// Create a memory copy of what's currently on the screen
		hDCScreen = GetDC(hWnd);
	    GetClientRect(hWnd, &rcWnd);
		hDCMem = CreateCompatibleDC(hDCScreen);
		hbmImage = CreateCompatibleBitmap(hDCScreen, rcWnd.right, rcWnd.bottom);
		hOldBitmap = SelectObject(hDCMem, hbmImage);
		BitBlt(hDCMem, 0, 0, rcWnd.right, rcWnd.bottom,  hDCScreen, 0, 0, SRCCOPY);
		
		/* convert the bitmap to a DIB */
		pdib = DibFromBitmap(hbmImage, 0, 0, NULL, NULL);
	
		if (((ToonIsTopDown (hToon)) && (pdib->biHeight > 0)) ||
			((!ToonIsTopDown (hToon)) && (pdib->biHeight < 0)))
		DibFlip(pdib);
		ToonSetBackground(hToon, pdib, DibPtr(pdib));
		ToonSetColors(hToon, Colors);
		ToonRestoreBackground(hToon);
		InvalidateRect(hWnd, NULL, FALSE);
	
		SelectObject(hDCMem, hOldBitmap);
		ReleaseDC(hWnd, hDCScreen);
	
		/* clean up */
		if(hbmImage)
			DeleteObject(hbmImage);
	
		if(hDCMem)
			DeleteDC(hDCMem);
	
		if (hMovie)
			mciClose(hMovie, 0, &mciGeneric);
	    
		if ((hMovie = ToonOpenMovie(hToon, szMovie, iTransparent)) != NULL)
		{
			MCI_DGV_RECT_PARMS mciRect;
						
			ErrCode = mciSendCommand (hMovie, MCI_WHERE, MCI_WAIT | MCI_DGV_WHERE_SOURCE, (DWORD)(LPMCI_DGV_RECT_PARMS) &mciRect);
			if (ErrCode)
			{
				mciGetErrorString(ErrCode, szBuf, sizeof(szBuf));
				MessageBox(hWnd, szBuf, "Couldn't set source ", MB_ICONEXCLAMATION);
				return FALSE;
			}
						
			ErrCode = mciSendCommand (hMovie, MCI_PUT, MCI_WAIT | MCI_DGV_PUT_DESTINATION | MCI_DGV_RECT, (DWORD)(LPMCI_DGV_RECT_PARMS) &mciRect);
			if (ErrCode)
			{
				mciGetErrorString(ErrCode, szBuf, sizeof(szBuf));
				MessageBox(hWnd, szBuf, "Couldn't set destination", MB_ICONEXCLAMATION);
				return FALSE;
			}
						
			ToonSetPaintList(hToon, LIST_NORMAL);
			ToonSetPaintMethod(hToon, PAINT_INVALIDATE_RECT);
	
			if(!bSound) 		// Turn the sound off if the user specifies this
			{
				ErrCode = mciSendCommand (hMovie, MCI_SETAUDIO, MCI_SET_OFF, NULL);
				if (ErrCode)
				{
					mciGetErrorString(ErrCode, szBuf, sizeof(szBuf));
					MessageBox(hWnd, szBuf, "Couldn't turn off audio", MB_ICONEXCLAMATION);
					return FALSE;
				}
	        }
	        else
			{
				ErrCode = mciSendCommand (hMovie, MCI_SETAUDIO, MCI_SET_ON, NULL);
				if (ErrCode)
				{
					mciGetErrorString(ErrCode, szBuf, sizeof(szBuf));
					MessageBox(hWnd, szBuf, "Couldn't turn on audio", MB_ICONEXCLAMATION);
					return FALSE;
				}
	        }
	
			// Start the movie
			OnPlay(hWnd);
	
	
			// Create a timer which is used to restart the movie
			wTimer = SetTimer(hWnd, ID_TIMER, wElapse, NULL);
	
			return TRUE;
		}
		else
		{
			if (pdib)
				DibFree(pdib);
			return FALSE;
		}
	}
	else
		return FALSE;	
}


/* OnToonDraw -- handle WM_TOONDRAW message
 *
 * Params:  handle to window, handle to toon, frame counter
 *			
 *
 * Return:  None
 */
void OnToonDraw(HWND hWnd, HTOON hToon, LONG lFrame)
{
	// First time through, force a draw of the background
	if(fRedrawBackground)
	{
		fRedrawBackground = FALSE;
		ToonRestoreBackground(hToon);
		InvalidateRect(hWnd, NULL, FALSE);
	}
	// Put any custom draws, which should be under the toon, here
	ToonDrawCurrentFrame(hToon);
	// Put any custom draws, which should be over the toon, here
}


